from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from pathlib import Path
from typing import Callable, Mapping, MutableMapping, Optional

import jsonschema

from cl_sii.libs.json_utils import JsonSchemaValidationError, read_json_schema
from cl_sii.rcv.data_models import PeriodoTributario
from cl_sii.rut import Rut
from .data_models import CteForm29


SiiCteF29DatosObjType = Mapping[str, Mapping[str, object]]
_CTE_F29_DATOS_OBJ_SCHEMA_PATH = (
    Path(__file__).parent.parent.parent
    / 'data'
    / 'cte'
    / 'schemas-json'
    / 'f29_datos_obj.schema.json'
)
CTE_F29_DATOS_OBJ_SCHEMA = read_json_schema(_CTE_F29_DATOS_OBJ_SCHEMA_PATH)


def parse_sii_cte_f29_datos_obj(
    datos_obj: SiiCteF29DatosObjType,
    schema_validator: Optional[Callable[[SiiCteF29DatosObjType], None]] = None,
    campo_deserializer: Optional[Callable[[object, str], object]] = None,
) -> CteForm29:
    """
    Parse the ``datos`` JavaScript object embedded in the IFrames of the
    HTML version of the CTE's 'Declaraciones de IVA (F29)'.

    :param schema_validator: Schema validator. For the signature of the callable, see the docstring
        of ``cte_f29_datos_schema_default_validator``.
    :param campo_deserializer: 'Campo' deserializer. For the signature of the callable, see the
        docstring of ``cte_f29_datos_obj_campo_default_deserializer``.
    :raises JsonSchemaValidationError: If schema validation fails.
    """
    if schema_validator is None:
        schema_validator = cte_f29_datos_schema_default_validator
    if campo_deserializer is None:
        campo_deserializer = cte_f29_datos_obj_campo_default_deserializer

    obj_params = _parse_sii_cte_f29_datos_obj_to_dict(
        datos_obj=datos_obj,
        schema_validator=schema_validator,
        campo_deserializer=campo_deserializer,
    )
    obj = CteForm29(**obj_params)
    return obj


def _parse_sii_cte_f29_datos_obj_to_dict(
    datos_obj: SiiCteF29DatosObjType,
    schema_validator: Callable[[SiiCteF29DatosObjType], None],
    campo_deserializer: Callable[[object, str], object],
) -> Mapping[str, object]:
    """
    Parse the ``datos`` JavaScript object embedded in the IFrames of the
    HTML version of the CTE's 'Declaraciones de IVA (F29)' and return a
    dictionary.

    :param schema_validator:
    :param campo_deserializer:
    :raises JsonSchemaValidationError: If schema validation fails.
    """
    schema_validator(datos_obj)

    datos_obj_campos: Mapping[int, str] = {
        int(code): str(value) for code, value in datos_obj['campos'].items()
    }
    datos_obj_extras: Mapping[str, object] = datos_obj['extras']
    datos_obj_glosa: Mapping[int, str] = {  # noqa: F841
        int(code): str(value) for code, value in datos_obj['glosa'].items()
    }
    datos_obj_tipos: Mapping[int, str] = {
        int(code): str(value) for code, value in datos_obj['tipos'].items()
    }

    deserialized_datos_obj_campos = {
        code: campo_deserializer(value, datos_obj_tipos[code])
        for code, value in datos_obj_campos.items()
    }

    obj_extra: MutableMapping[int, object] = {}
    for code, value in deserialized_datos_obj_campos.items():
        obj_extra[code] = value

    periodo_tributario = PeriodoTributario.from_date(
        datetime.strptime(str(datos_obj_extras['PERIODO']), '%Y%m').date(),
    )

    obj_dict = dict(
        folio=obj_extra[7],
        contribuyente_rut=obj_extra[3],
        periodo_tributario=periodo_tributario,
        #
        tipo_declaracion=datos_obj_extras.get('CLASE'),
        banco=datos_obj_extras.get('BANCO'),
        medio_pago=datos_obj_extras.get('MEDIO_PAGO'),
        #
        extra=obj_extra,
    )
    return obj_dict


def cte_f29_datos_obj_campo_default_deserializer(campo_value: object, tipo: str) -> object:
    """
    Convert raw values from the ``datos`` object to the corresponding Python
    data type.

    :param campo_value: Raw value from 'campos'.
    :param tipo: Data type code from 'tipos'.
    :raises ValueError: If value conversion fails.
    """
    if not isinstance(campo_value, str):
        return campo_value

    deserialized_value: object

    if tipo == 'R':
        deserialized_value = Rut(campo_value)
    elif tipo == 'F':
        deserialized_value = datetime.strptime(campo_value, '%d/%m/%Y').date()
    elif tipo == 'C':
        deserialized_value = campo_value
    elif tipo in ('N', 'M'):
        deserialized_value = int(campo_value)
    elif tipo == 'D':
        deserialized_value = Decimal(campo_value)
    else:
        raise ValueError(f"Invalid or unknown tipo of campo: '{tipo}'")

    return deserialized_value


def cte_f29_datos_obj_campo_best_effort_deserializer(campo_value: object, tipo: str) -> object:
    """
    Convert raw values from the ``datos`` object to the corresponding Python
    data type. Values that cannot be converted will be returned without modification.

    :param campo_value: Raw value from 'campos'.
    :param tipo: Data type code from 'tipos'.
    """
    try:
        deserialized_value = cte_f29_datos_obj_campo_default_deserializer(
            campo_value=campo_value,
            tipo=tipo,
        )
    except Exception:
        deserialized_value = campo_value

    return deserialized_value


def cte_f29_datos_schema_default_validator(datos_obj: SiiCteF29DatosObjType) -> None:
    """
    Validate the ``datos`` object against the schema.

    :raises JsonSchemaValidationError: If schema validation fails.
    :returns: ``None`` if schema validation passed.
    """
    try:
        jsonschema.validate(datos_obj, schema=CTE_F29_DATOS_OBJ_SCHEMA)
    except jsonschema.exceptions.ValidationError as exc:
        raise JsonSchemaValidationError('Validation against JSON Schema failed') from exc

    if datos_obj['campos'].keys() != datos_obj['tipos'].keys():
        raise JsonSchemaValidationError("The keys of 'campos' and 'tipos' are not exactly the same")
    if datos_obj['campos'].keys() != datos_obj['glosa'].keys():
        raise JsonSchemaValidationError("The keys of 'campos' and 'tipos' are not exactly the same")



import enum
from typing import FrozenSet

from cl_sii.dte.constants import DTE_MONTO_TOTAL_FIELD_MAX_VALUE, TipoDte


# The collection of "tipo DTE" for which it is possible to "ceder" a "DTE".
#   They are defined in a document and also an XML schema.
#   - Document "Formato Archivo Electrónico de Cesión (AEC)"
#     (http://www.sii.cl/factura_electronica/cesion.pdf) are:
#     > Sólo códigos 33, 34, 46 y 43
#   - XML element 'CesionDefType/DocumentoCesion/IdDTE/TipoDTE'
#     - description: "Tipo de DTE"
#     - XML type: 'SiiDte:DTEFacturasType'
#     - source:
#       https://github.com/fyntex/lib-cl-sii-python/blob/7e1c4b52/cl_sii/data/ref/factura_electronica/schemas-xml/Cesion_v10.xsd#L38-L42
#   - XML type 'SiiDte:DTEFacturasType' in official schema 'SiiTypes_v10.xsd'
#     - source:
#       https://github.com/fyntex/lib-cl-sii-python/blob/7e1c4b52/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd#L100-L126
TIPO_DTE_CEDIBLES: FrozenSet[TipoDte] = frozenset(
    {
        TipoDte.FACTURA_ELECTRONICA,
        TipoDte.FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA,
        TipoDte.FACTURA_COMPRA_ELECTRONICA,
        TipoDte.LIQUIDACION_FACTURA_ELECTRONICA,
    }
)


###############################################################################
# Cesion Fields / "Monto Cedido"
###############################################################################

# Amount of the "cesión".
#
# Ref:
#   - https://github.com/fyntex/lib-cl-sii-api-python/blob/v0.4.4/cl_sii_api/rtc/data_models.py#L231
#   - Document "Formato Archivo Electrónico de Cesión 2013-02-11" (retrieved on 2019-08-12)
#     (https://www.sii.cl/factura_electronica/cesion.pdf)
CESION_MONTO_CEDIDO_FIELD_MIN_VALUE: int = 0
CESION_MONTO_CEDIDO_FIELD_MAX_VALUE: int = DTE_MONTO_TOTAL_FIELD_MAX_VALUE


###############################################################################
# Cesion Fields / "Secuencia"
###############################################################################

# Sequence number of the "cesión"
#
# > Campo: Número de Cesión
# > Descripción: Secuencia de la cesión
# > Tipo: NUM
# > Validación: 1 hasta 40
#
# Source:
#   Document "Formato Archivo Electrónico de Cesión 2013-02-11" (retrieved on 2019-08-12)
#   (https://www.sii.cl/factura_electronica/cesion.pdf)
CESION_SEQUENCE_NUMBER_MIN_VALUE: int = 1
CESION_SEQUENCE_NUMBER_MAX_VALUE: int = 40


###############################################################################
# Other
###############################################################################


@enum.unique
class RolContribuyenteEnCesion(enum.Enum):

    """
    "Rol" of "contribuyente" in a "cesion".
    """

    CEDENTE = 'CEDENTE'
    """Cesiones en las que el contribuyente ha sido cedente i.e. ha cedido"""

    CESIONARIO = 'CESIONARIO'
    """Cesiones en las que el contribuyente ha sido cesionario i.e. le han cedido"""

    DEUDOR = 'DEUDOR'
    """Cesiones de DTEs en que el contribuyente es el deudor."""



import re
from typing import Optional

import cryptography
import cryptography.x509
from cryptography.hazmat.backends.openssl import backend as crypto_x509_backend

from . import Rut, constants


def get_subject_rut_from_certificate_pfx(pfx_file_bytes: bytes, password: Optional[str]) -> Rut:
    """
    Return the Chilean RUT stored in a digital certificate.

    Original source URL: https://github.com/fyntex/fd-cl-data/blob/cfd5a716fb9b2cbd8a03fca1bacfd1b844b1337f/fd_cl_data/apps/sii_auth/models/sii_auth_credential.py#L701-L745  # noqa: E501

    :param pfx_file_bytes: Digital certificate in PKCS12 format
    :param password: (Optional) The password to use to decrypt the PKCS12 file
    """
    (
        private_key,
        x509_cert,
        additional_certs,
    ) = crypto_x509_backend.load_key_and_certificates_from_pkcs12(
        data=pfx_file_bytes,
        password=password.encode() if password is not None else None,
    )
    # https://cryptography.io/en/latest/hazmat/primitives/asymmetric/serialization/#cryptography.hazmat.primitives.serialization.pkcs12.load_key_and_certificates  # noqa: E501

    subject_alt_name_ext = x509_cert.extensions.get_extension_for_class(
        cryptography.x509.extensions.SubjectAlternativeName,
    )

    # Search for the RUT in the certificate.
    try:
        results = [
            x.value
            for x in subject_alt_name_ext.value._general_names
            if x.type_id == constants.SII_CERT_TITULAR_RUT_OID
        ]
    except AttributeError:
        raise Exception('Certificate has no RUT information')

    if not results:
        raise Exception('Certificate has no RUT information')
    elif len(results) > 1:
        raise Exception(f'len(results) == {len(results)}')

    subject_rut_raw: bytes = results[0]
    subject_rut = re.sub(r'[^0-9-]', '', subject_rut_raw.decode('utf-8'))

    return Rut(subject_rut)
